(* elements of calculations.
   they are partially held in association lists as ref's for
   switching language levels (meta-string, object-values).
   in order to keep these ref's during re-evaluation of code,
   they are defined here at the beginning of the code.
   author: Walther Neuper
   (c) isac-team 2003
 
use"calcelems.sml";
*)

val linefeed = (curry op^) "\n";
type authors = string list;

type cterm' = string;
val empty_cterm' = "empty_cterm'";
type thmID = string;
type thm' = thmID * cterm';(*WN060610 deprecated in favour of thm''*)
type thm'' = thmID * term;
type rls' = string;
(*.a 'guh'='globally unique handle' is a string unique for each element 
   of isac's KEStore and persistent over time
   (in particular under shifts within the respective hierarchy);
   specialty for thys: 
   # guh NOT resistant agains shifts from one thy to another
   (which is the price for Isabelle's design: thy's overwrite ids of subthy's)
   # requirement for matchTheory: induce guh from tac + current thy
   (see 'fun thy_containing_thm', 'fun thy_containing_rls' etc.)
   TODO: introduce to pbl, met.*) 
type guh = string;
val e_guh = "e_guh":guh;

type xml = string;

(*.fun calculate_ fetches the evaluation-function via this list.*)
(*ie. there is NO rewriting involved in contrary to eval_rls*)
type calc =
     (string *     (*op in calculate_ from scripts*)
      (string *    (*op in isa-term 'Const(op,_)' *)
       (string -> term -> theory -> (string * term) option))); (*eval_fun*) 

fun e_evalfn (_:'a) (_:term) (_:theory) = None:(string * term) option;

type subs' = (cterm' * cterm') list; (*16.11.00 for FE-KE*)
type subst = (term * term) list; (*here for ets2str*)
val e_subst = []:(term * term) list;

(*TODO.WN060610 make use of "type rew_ord" total*)
type rew_ord' = string;
fun dummy_ord (_:subst) (_:term,_:term) = true;
type rew_ord = rew_ord' * (subst -> Term.term * Term.term -> bool);
val e_rew_ord = dummy_ord;

datatype rule = 
  Erule (*WN.3.6.03 for rep_tac_ Check_elementwise*)
| Thm of (string * thm) 
| Calc of string *      (*check for equality*)
	  (string -> term -> theory -> (string * term) option)
| Rls_ of rls         (*.ie. rule sets may be nested.*)
and scr = 
    EmptyScr 
  | Script of term (*for met*)
  | Rfuns of {init_state : term -> 
     (term *          (*the current formula: 
                        goes locate_gen -> next_tac via istate*)
      term *          (*the final formula*)
      rule list       (*of reverse rewrite set (#1#)*)
	    list *    (*may be serveral, eg. in norm_rational*)
      (rule *         (*Thm (+ Thm generated from Calc) resulting in ...*)
       (term *        (*... rewrite with ...*)
	term list))   (*... assumptions*)
	  list),      (*derivation from given term to normalform
		       in reverse order with sym_thm; 
                       (#1#) could be extracted from here #1*)

	      normal_form: term -> (term * term list) option,
	      locate_rule: rule list list -> term -> rule 
			   -> (rule * (term * term list)) list,
	      next_rule  : rule list list -> term -> rule option,
	      attach_form: rule list list -> term -> term 
			   -> (rule * (term * term list)) list}
and rls =
    Erls                          (*for init e_rls*)
  
  | Rls of (*a confluent and terminating ruleset, in general         *)
    {id : string,          (*for trace_rewrite:=true                 *)
     preconds : term list, (*unused WN020820                         *)
     (*WN060616 for efficiency...
      bdvs    : false,       (*set in prep_rls for get_bdvs *)*)
     rew_ord  : rew_ord,   (*for rules*)
     erls     : rls,       (*for the conditions in rules             *)
     srls     : rls,       (*for evaluation of list_fns in script    *)
     calc     : calc list, (*for Calculate in scr, set by prep_rls   *)
     rules    : rule list,
     scr      : scr}       (*Script term: generating intermed.steps  *)
  | Seq of (*a sequence of rules to be tried only once               *)
    {id : string,          (*for trace_rewrite:=true                 *)
     preconds : term list, (*unused 20.8.02                          *)
     (*WN060616 for efficiency...
      bdvs    : false,       (*set in prep_rls for get_bdvs *)*)
     rew_ord  : rew_ord,   (*for rules                               *)
     erls     : rls,       (*for the conditions in rules             *)
     srls     : rls,       (*for evaluation of list_fns in script    *)
     calc     : calc list, (*for Calculate in scr, set by prep_rls   *) 
     rules    : rule list,
     scr      : scr}  (*Script term  (how to restrict type ???)*)
  (*Rrls call SML-code and simulate an rls
    difference: there is always _ONE_ redex rewritten in 1 call,
    thus wrap Rrls by: Rls (Rls_ ...)*)
  
  | Rrls of (*for 'reverse rewriting' by SML-functions instead Script*)
    {id : string,          (*for trace_rewrite:=true                 *)
     prepat  : (term list *(*preconds, eval with subst from pattern  *)
		term )     (*pattern matched in subterms             *)
		   list,   (*meta-conjunction is or                  *)
     rew_ord  : rew_ord,   (*for rules                               *)
     erls     : rls,       (*for the conditions in rules and pat     *)
     (*            '^ because of rewrite in applicable_in
						compare type met*)
     calc     : calc list, (*for Calculate in scr, set by prep_rls *)
     scr      : scr}; (*Rfuns {...}  (how to restrict type ???)*)
(*1.8.02 ad (how to restrict type ???): scr should be usable indepentently
  from rls, and then contain both Script _AND_ Rfuns !!!*)

val e_rule = Thm ("refl",refl) : rule;
fun id_of_thm (Thm (id, _)) = id
  | id_of_thm _ = raise error "id_of_thm";
fun thm_of_thm (Thm (_, thm)) = thm
  | thm_of_thm _ = raise error "thm_of_thm";
fun rep_thm_G' (Thm (thmid, thm)) = (thmid, thm);
fun eq_thmI ((thmid1 : thmID, _ : thm), (thmid2 : thmID, _ : thm)) =
    (strip_thy thmid1) = (strip_thy thmid2);

(*.id requested for all, Rls,Seq,Rrls.*)
fun id_rls Erls = "e_rls" (*WN060714 quick and dirty: recursive defs!*)
  | id_rls (Rls {id,...}) = id
  | id_rls (Seq {id,...}) = id
  | id_rls (Rrls {id,...}) = id;
val rls2str = id_rls;
fun id_rule (Thm (id, _)) = id
  | id_rule (Calc (id, _)) = id
  | id_rule (Rls_ rls) = id_rls rls;

fun get_rules (Rls {rules,...}) = rules
  | get_rules (Seq {rules,...}) = rules
  | get_rules (Rrls _) = [];

fun rule2str Erule = "Erule"
  | rule2str (Thm (str, thm)) = "Thm (\""^str^"\","^(string_of_thm thm)^")"
  | rule2str (Calc (str,f))  = "Calc (\""^str^"\",fn)"
  | rule2str (Rls_ rls) = "Rls_ (\""^id_rls rls^"\")";
fun rule2str' Erule = "Erule"
  | rule2str' (Thm (str, thm)) = "Thm (\""^str^"\",\"\")"
  | rule2str' (Calc (str,f))  = "Calc (\""^str^"\",fn)"
  | rule2str' (Rls_ rls) = "Rls_ (\""^id_rls rls^"\")";

fun eqrule (Thm (id1,_), Thm (id2,_)) = id1 = id2
  | eqrule (Calc (id1,_), Calc (id2,_)) = id1 = id2
  | eqrule (Rls_ _, Rls_ _) = false (*{id=id1}{id=id2} = id1 = id2 FIXXME*)
  | eqrule _ = false;


type rrlsstate =      (*state for reverse rewriting*)
     (term *          (*the current formula: 
                        goes locate_gen -> next_tac via istate*)
      term *          (*the final formula*)
      rule list       (*of reverse rewrite set (#1#)*)
	    list *    (*may be serveral, eg. in norm_rational*)
      (rule *         (*Thm (+ Thm generated from Calc) resulting in ...*)
       (term *        (*... rewrite with ...*)
	term list))   (*... assumptions*)
	  list);      (*derivation from given term to normalform
		       in reverse order with sym_thm; 
                       (#1#) could be extracted from here #1*)
val e_type = Type("empty",[]);
val a_type = TFree("'a",[]);
val e_term = Const("empty",e_type);
val a_term = Free("empty",a_type);
val e_rrlsstate = (e_term,e_term,[[e_rule]],[(e_rule,(e_term,[]))]):rrlsstate;




(*22.2.02: ging auf Linux nicht (Stefan)
val e_scr = Script ((term_of o the o (parse thy)) "e_script");*)
val e_term = Const("empty", Type("'a", []));
val e_scr = Script e_term;


(*ad thm':
   there are two kinds of theorems ...
   (1) known by isabelle
   (2) not known, eg. calc_thm, instantiated rls
       the latter have a thmid "#..."
   and thus outside isa we ALWAYS transport both(thmid,string_of_thm)
   and have a special assoc_thm / assoc_rls in this interface      *)
type theory' = string; (* = domID ^".thy" *)
type domID = string;   (* domID ^".thy" = theory' TODO.11.03replace by thyID*)
type thyID = string;    (*WN.3.11.03 TODO: replace domID with thyID*)

fun string_of_thy thy =  
((last_elem (Sign.stamp_names_of (sign_of thy)))^".thy"):theory';
val theory2domID = string_of_thy;
val theory2thyID = (get_thy o string_of_thy) : theory -> thyID;
val theory2theory' = string_of_thy;
val theory2str = string_of_thy; (*WN050903 ..most consistent naming*)
val theory2str' = implode o (drop_last_n 4) o explode o string_of_thy;
(*> theory2str' Isac.thy;
al it = "Isac" : string
*)

fun thyID2theory' (thyID:thyID) =
    let val ss = explode thyID
	val ext = implode (takelast (4, ss))
    in if ext = ".thy" then thyID : theory' (*disarm abuse of thyID*)
       else thyID ^ ".thy"
    end;
(* thyID2theory' "Isac" (*ok*);
val it = "Isac.thy" : theory'
 > thyID2theory' "Isac.thy" (*abuse, goes ok...*);
val it = "Isac.thy" : theory'
*)

fun theory'2thyID (theory':theory') =
    let val ss = explode theory'
	val ext = implode (takelast (4, ss))
    in if ext = ".thy" then ((implode o (drop_last_n 4)) ss) : thyID
       else theory' (*disarm abuse of theory'*)
    end;
(* theory'2thyID "Isac.thy";
val it = "Isac" : thyID
> theory'2thyID "Isac";
val it = "Isac" : thyID*)


(*. WN0509 discussion:
#############################################################################
#   How to manage theorys in subproblems wrt. the requirement,              #
#   that scripts should be re-usable ?                                      #
#############################################################################

    eg. 'Script Solve_rat_equation' calls 'SubProblem (RatEq_,..'
    which would not allow to 'solve (y'' = -M_b / EI, M_b)' by this script
    because Biegelinie.thy is subthy of RatEq.thy and thus Biegelinie.M_b 
    is unknown in RatEq.thy and M_b cannot be parsed into the scripts guard
    (see match_ags).

    Preliminary solution:
    # the thy in 'SubProblem (thy_, pbl, arglist)' is not taken automatically,
    # instead the 'maxthy (rootthy pt) thy_' is taken for each subpbl
    # however, a thy specified by the user in the rootpbl may lead to 
      errors in far-off subpbls (which are not yet reported properly !!!) 
      and interactively specifiying thys in subpbl is not very relevant.

    Other solutions possible:
    # always parse and type-check with Isac.thy
      (rejected tue to the vague idea eg. to re-use equations for R in C etc.)
    # regard the subthy-relation in specifying thys of subpbls
    # specifically handle 'SubProblem (undefined_, pbl, arglist)'
    # ???
.*)
(*WN0509 TODO "ProtoPure" ... would be more consistent 
  with assoc_thy <--> theory2theory' +FIXME assoc_thy "e_domID" -> Script.thy*)
val e_domID = "e_domID":domID;

(*the key into the hierarchy ob theory elements*)
type theID = string list;
val e_theID = ["e_theID"];
val theID2str = strs2str;
(*theID eg. is ["IsacKnowledge", "Test", "Rulesets", "ac_plus_times"]*)
fun theID2thyID (theID:theID) =
    if length theID >= 3 then (last_elem o (drop_last_n 2)) theID : thyID
    else raise error ("theID2thyID called with "^ theID2str theID);

(*the key into the hierarchy ob problems*)
type pblID = string list; (* domID::...*)
val e_pblID = ["e_pblID"]:pblID;
val pblID2str = strs2str;

(*the key into the hierarchy ob methods*)
type metID = string list;
val e_metID = ["e_metID"]:metID;
val metID2str = strs2str;

(*either theID or pblID or metID*)
type kestoreID = string list;
val e_kestoreID = ["e_kestoreID"];
val kestoreID2str = strs2str;

(*for distinction of contexts*)
datatype ketype = Exp_ | Thy_ | Pbl_ | Met_;
fun ketype2str Exp_ = "Exp_"
  | ketype2str Thy_ = "Thy_" 
  | ketype2str Pbl_ = "Pbl_" 
  | ketype2str Met_ = "Met_";
fun ketype2str' Exp_ = "Example"
  | ketype2str' Thy_ = "Theory" 
  | ketype2str' Pbl_ = "Problem" 
  | ketype2str' Met_ = "Method";

(*see 'How to manage theorys in subproblems' at 'type thyID'*)
val theory'  = ref ([]:(theory' * theory) list);

(*.all theories defined for Scripts, recorded in Scripts/Script.ML; 
   in order to distinguish them from general IsacKnowledge defined later on.*)
val script_thys = ref ([] : (theory' * theory) list);


(*rewrite orders, also stored in 'type met' and type 'and rls'
  The association list is required for 'rewrite.."rew_ord"..'
  WN0509 tests not well-organized: see smltest/IsacKnowledge/termorder.sml*)
val rew_ord' = 
    ref ([]:(rew_ord' *        (*the key for the association list         *)
	     (subst 	       (*the bound variables - they get high order*)
	      -> (term * term) (*(t1, t2) to be compared                  *)
	      -> bool))        (*if t1 <= t2 then true else false         *)
		list);         (*association list                         *)
rew_ord' := overwritel (!rew_ord', [("e_rew_ord", e_rew_ord),
				    ("dummy_ord", dummy_ord)]);


(*WN060120 a hack to get alltogether run again with minimal effort:
  theory' is inserted for creating thy_hierarchy; calls for assoc_rls
  need not be called*)
val ruleset' = ref ([]:(rls' * (theory' * rls)) list);

(*FIXME.040207 calclist': used by prep_rls, NOT in met*)
val calclist'= ref ([]: calc list);

(*.the hierarchy of thydata.*)

(*.'a is for pbt | met.*)
(*WN.24.4.03 -"- ... type parameters; afterwards naming inconsistent*)
datatype 'a ptyp = 
	 Ptyp of string *   (*element within pblID*)
		 'a list *  (*several pbts with different domIDs/thy
			      TODO: select by subthy (isaref.p.69)
			      presently only _ONE_ elem*)
		 ('a ptyp) list;   (*the children nodes*)

(*.datatype for collecting thydata for hierarchy.*)
(*WN060720 more consistent naming would be 'type thyelem' or 'thelem'*)
(*WN0606 Htxt contains html which does not belong to the sml-kernel*)
datatype thydata = Html of {guh: guh,
			    coursedesign: authors,
			    mathauthors: authors,
			    html: string} (*html; for demos before database*)
		 | Hthm of {guh: guh,
			    coursedesign: authors,
			    mathauthors: authors,
			    thm: Thm.thm}
		 | Hrls of {guh: guh,
			    coursedesign: authors,
			    mathauthors: authors,
			    (*like   vvvvvvvvvvvvv val ruleset'
			     WN060711 redesign together !*)
			    thy_rls: (thyID * rls)}
		 | Hcal of {guh: guh,
			    coursedesign: authors,
			    mathauthors: authors,
			    calc: calc}
		 | Hord of {guh: guh,
			    coursedesign: authors,
			    mathauthors: authors,
			    ord: (subst -> (term * term) -> bool)};
val e_thydata = Html {guh="e_guh", coursedesign=[], mathauthors=[], html=""};

type thehier = (thydata ptyp) list;
val thehier = ref ([] : thehier);

(*.an association list, gets the value once in Isac.ML.*)
val isab_thm_thy = ref ([] : (thmID * (thyID * thm)) list);


type path = string;
type filename = string;

(*val xxx = fn: a b => (a,b);   ??? fun-definition ???*)
local
    fun ii (_:term) = e_rrlsstate;
    fun no (_:term) = Some (e_term,[e_term]);
    fun lo (_:rule list list) (_:term) (_:rule) = [(e_rule,(e_term,[e_term]))];
    fun ne (_:rule list list) (_:term) = Some e_rule;
    fun fo (_:rule list list) (_:term) (_:term) = [(e_rule,(e_term,[e_term]))];
in
val e_rfuns = Rfuns {init_state=ii,normal_form=no,locate_rule=lo,
		     next_rule=ne,attach_form=fo};
end;

val e_rls = 
  Rls{id = "e_rls",
      preconds = [], 
      rew_ord = ("dummy_ord", dummy_ord),
      erls = Erls,srls = Erls,
      calc = [],
      rules = [], scr = EmptyScr}:rls;
val e_rrls = Rrls {id = "e_rrls",
		   prepat = [],
		   rew_ord = ("dummy_ord", dummy_ord),
		   erls = Erls,
		   calc = [],
		   (*asm_thm=[],*)
		   scr=e_rfuns}:rls;
ruleset' := overwritel (!ruleset', [("e_rls",("Tools",e_rls)),
				    ("e_rrls",("Tools",e_rrls))
				    ]);

fun rep_rls (Rls {id,preconds,rew_ord,erls,srls,calc,(*asm_thm,*)rules,scr}) =
  {id=id,preconds=preconds,rew_ord=rew_ord,erls=erls,srls=srls,calc=calc,
   (*asm_thm=asm_thm,*)rules=rules,scr=scr}
  | rep_rls (Seq {id,preconds,rew_ord,erls,srls,calc,(*asm_thm,*)rules,scr}) =
  {id=id,preconds=preconds,rew_ord=rew_ord,erls=erls,srls=srls,calc=calc,
   (*asm_thm=asm_thm,*)rules=rules,scr=scr}
  | rep_rls Erls = rep_rls e_rls
  | rep_rls (Rrls {id,...})  = rep_rls e_rls
    (*raise error("rep_rls doesn't take apart reverse-rewrite-rule-sets: "^id)*);
(*| rep_rls (Seq {id,...})  = 
    raise error("rep_rls doesn't take apart reverse-rewrite-rule-sets: "^id);
--1.7.03*)
fun rep_rrls 
	(Rrls {id,(*asm_thm,*) calc, erls, prepat, rew_ord, 
	       scr=Rfuns
		       {attach_form,init_state,locate_rule,
			next_rule,normal_form}}) =
    {id=id,(*asm_thm=asm_thm,*) calc=calc, erls=erls, prepat=prepat, 
     rew_ord=rew_ord, attach_form=attach_form, init_state=init_state, 
     locate_rule=locate_rule, next_rule=next_rule, normal_form=normal_form}
  | rep_rrls (Rls {id,...}) = 
    raise error ("rep_rrls doesn't take apart (normal) rule-sets: "^id)
  | rep_rrls (Seq {id,...}) = 
    raise error ("rep_rrls doesn't take apart (normal) rule-sets: "^id);

fun append_rls id (Rls {id=_,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
			rules =rs,scr=sc}) r =
    (Rls{id=id,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
	 rules = rs @ r,scr=sc}:rls)
  | append_rls id (Seq {id=_,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
			rules =rs,scr=sc}) r =
    (Seq{id=id,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
	 rules = rs @ r,scr=sc}:rls)
  | append_rls id (Rrls _) _ = 
    raise error ("append_rls: not for reverse-rewrite-rule-set "^id);

fun eq_rule (Thm (thm1,_), Thm (thm2,_)) = thm1 = thm2
  | eq_rule (Calc (id1,_), Calc (id2,_)) = id1 = id2
  | eq_rule _ = false;

fun merge_rls _ Erls rls = rls
  | merge_rls _ rls Erls = rls
  | merge_rls id
	(Rls {id=id1,preconds=pc1,rew_ord=ro1,erls=er1,srls=sr1,calc=ca1,
	      (*asm_thm=at1,*)rules =rs1,scr=sc1}) 
	(r2 as Rls {id=id2,preconds=pc2,rew_ord=ro2,erls=er2,srls=sr2,calc=ca2,
	      (*asm_thm=at2,*)rules =rs2,scr=sc2}) =
	(Rls {id=id,preconds=pc1 @ ((#preconds o rep_rls) r2),
	      rew_ord=ro1,erls=merge_rls "" er1 er2(*er1*),
	      srls=merge_rls ("merged_"^id1^"_"^((#id o rep_rls) r2)) sr1 
			     ((#srls o rep_rls) r2),
	      calc=ca1 @ ((#calc o rep_rls) r2),
	      (*asm_thm=at1 @ ((#asm_thm o rep_rls) r2),*)
	      rules = gen_union eq_rule rule2str (rs1, (#rules o rep_rls) r2),
	      scr=sc1}:rls)
  | merge_rls id
	(Seq {id=id1,preconds=pc1,rew_ord=ro1,erls=er1,srls=sr1,calc=ca1,
	      (*asm_thm=at1,*)rules =rs1,scr=sc1}) 
	(r2 as Seq {id=id2,preconds=pc2,rew_ord=ro2,erls=er2,srls=sr2,calc=ca2,
	      (*asm_thm=at2,*)rules =rs2,scr=sc2}) =
	(Seq {id=id,preconds=pc1 @ ((#preconds o rep_rls) r2),
	      rew_ord=ro1,erls=merge_rls "" er1 er2(*er1*),
	      srls=merge_rls ("merged_"^id1^"_"^((#id o rep_rls) r2)) sr1 
			     ((#srls o rep_rls) r2),
	      calc=ca1 @ ((#calc o rep_rls) r2),
	      (*asm_thm=at1 @ ((#asm_thm o rep_rls) r2),*)
	      rules = gen_union eq_rule rule2str (rs1, (#rules o rep_rls) r2),
	      scr=sc1}:rls)
  | merge_rls _ _ _ = 
    raise error "merge_rls: not for reverse-rewrite-rule-sets\
		\and not for mixed Rls -- Seq";
fun remove_rls id (Rls {id=_,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
		     (*asm_thm=at,*)rules =rs,scr=sc}) r =
    (Rls{id=id,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
	 (*asm_thm=at,*)rules = gen_rems eq_rule (rs, r),
	 scr=sc}:rls)
  | remove_rls id (Seq {id=_,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
		     (*asm_thm=at,*)rules =rs,scr=sc}) r =
    (Seq{id=id,preconds=pc,rew_ord=ro,erls=er,srls=sr,calc=ca,
	 (*asm_thm=at,*)rules = gen_rems eq_rule (rs, r),
	 scr=sc}:rls)
  | remove_rls id (Rrls _) _ = raise error 
                   ("remove_rls: not for reverse-rewrite-rule-set "^id);

(*!!!*)gen_rems (op=) ([1,2,3,4], [3,4,5]);

(*elder version...*)
fun mem_rls id rls = 
    case find_first ((curry op=) id) (map id_rule ((#rules o rep_rls) rls)) of
	Some _ => true | None => false;
fun memrls r (Rls {rules,...}) = gen_mem eqrule (r, rules)
  | memrls r (Seq {rules,...}) = gen_mem eqrule (r, rules)
  | memrls r _ = raise error ("memrls: incomplete impl. r= "^(rule2str r));

fun rls_get_thm rls id =
    case find_first (curry eq_rule (Thm (id, refl(*dummy*)))) 
		    ((#rules o rep_rls) rls) of
	Some thm => Some thm | None => None;



fun assoc' ([], key) = raise error ("ME_Isa: '"^key^"' not known")
  | assoc' ((keyi, xi) :: pairs, key) =
      if key = keyi then Some xi else assoc' (pairs, key);

fun assoc_thy (thy:theory') = ((the o assoc')(!theory',thy))
  handle _ => raise error ("ME_Isa: thy '"^thy^"' not in system");
(*.associate an rls-identifier with an rls; related to 'fun assoc_rls';
   these are NOT compatible to "fun assoc_thm'" in that they do NOT handle
   overlays by re-using an identifier in different thys.*)
fun assoc_rls (rls:rls') = ((#2 o the o assoc')(!ruleset',rls))
  handle _ => raise error ("ME_Isa: '"^rls^"' not in system");
(*fun assoc_rls (rls:rls') = ((the o assoc')(!ruleset',rls))
  handle _ => raise error ("ME_Isa: '"^rls^"' not in system");*)

(*.overwrite an element in an association list and pair it with a thyID
   in order to create the thy_hierarchy;
   overwrites existing rls' even if they are defined in a different thy;
   this is related to assoc_rls, TODO.WN060120: assoc_rew_ord, assoc_calc;.*)
(*WN060120 ...these are NOT compatible to "fun assoc_thm'" in that 
   they do NOT handle overlays by re-using an identifier in different thys;
   "thyID.rlsID" would be a good solution, if the "." would be possible
   in scripts...
   actually a hack to get alltogether run again with minimal effort*)
fun insthy thy' (rls', rls) = (rls', (thy', rls));
fun overwritelthy thy (al, bl:(rls' * rls) list) =
    let val bl' = map (insthy ((get_thy o theory2theory') thy)) bl
    in overwritel (al, bl') end;

fun assoc_rew_ord ro = ((the o assoc') (!rew_ord',ro))
  handle _ => raise error ("ME_Isa: rew_ord '"^ro^"' not in system");
(*get the string for stac from rule*)
fun assoc_calc ([], key) = raise error ("assoc_calc: '"^ key ^"' not found")
  | assoc_calc ((calc, (keyi, xi)) :: pairs, key) =
      if key = keyi then calc else assoc_calc (pairs, key);
(*only used for !calclist'*)
fun assoc1 ([], key) = raise error ("assoc1 (for met.calc=): '"^ key 
				    ^"' not found")
  | assoc1 ((all as (keyi, _)) :: pairs, key) =
      if key = keyi then all else assoc1 (pairs, key);
(*fun assoc_thm' requires mk_thm, num_str; see Scripts/rewrite.sml*)


fun termopt2str (Some t) = 
    "Some " ^ (Sign.string_of_term (sign_of(assoc_thy "Isac.thy")) t)
  | termopt2str None = "None";
fun term2str t = Sign.string_of_term (sign_of(assoc_thy "Isac.thy")) t;
(*for tests only*)
fun terms2str ts= (strs2str o (map (Sign.string_of_term 
					(sign_of (assoc_thy "Isac.thy"))))) ts;
fun terms2strs ts= ((map (Sign.string_of_term 
			      (sign_of (assoc_thy "Isac.thy"))))) ts;
fun type2str typ = Sign.string_of_typ (sign_of (assoc_thy "Isac.thy")) typ;

fun subst2str (s:subst) = 
    (strs2str o 
     (map (linefeed o pair2str o
	   (apsnd term2str) o 
	   (apfst term2str)))) s;
fun subst2str' (s:subst) = 
    (strs2str' o 
     (map (pair2str o
	   (apsnd term2str) o 
	   (apfst term2str)))) s;
(*> subst2str' [(str2term "bdv", str2term "x"),
		(str2term "bdv_2", str2term "y")];
val it = "[(bdv, x)]" : string
*)
val env2str = subst2str;


(*recursive defs:*)
fun scr2str (Script s) = "Script "^(term2str s)
  | scr2str (Rfuns _)  = "Rfuns";


fun maxthy thy1 thy2 = if subthy (thy1, thy2) then thy2 else thy1;


(*.trace internal steps of isac's rewriter*)
val trace_rewrite = ref false;
(*.depth of recursion in traces of the rewriter, if trace_rewrite:=true.*)
val depth = ref 99999;
(*.no of rewrites exceeding this int -> NO rewrite.*)
(*WN060829 still unused...*)
val lim_rewrite = ref 99999;
(*.no of derivation-elements exceeding this int -> SOME derivation-elements.*)
val lim_deriv = ref 100;
(*.switch for checking guhs unique before storing a pbl or met;
   set true at startup (done at begin of ROOT.ML)
   set false for editing IsacKnowledge (done at end of ROOT.ML).*)
val check_guhs_unique = ref false;


datatype lrd = (*elements of a path (=loc_) into an Isabelle term*)
	 L     (*go left at $*) 
       | R     (*go right at $*)
       | D;     (*go down at Abs*)
type loc_ = lrd list;
fun ldr2str L = "L"
  | ldr2str R = "R"
  | ldr2str D = "D";
fun loc_2str (k:loc_) = (strs2str' o (map ldr2str)) k;

